; Serial Packet decoder
; By Bill arden www.longrange.net
; 
;
;Entry
;	SerPktInit:
;	SerPktMain:
;	UART_RXC:
;	UART_DRE: 
;	UART_TXC: 
;Pins
;	PD0 = Rx
;	PD1 = Tx
;	PD5 = Select in
;	PD6 = Select out
;
;uses
;
;	REG_SerRXCsreg
;	REG_SerDREsreg
;	REG_SerTXCsreg
;

.DSEG
	SerAddress:		.byte 1		;[1] 0 = disabled
	SerGroup:		.byte 1		;[1] default = 255

	;SerBaud:		.byte 1		;[1] default = 63 (19200)
	SerStatusCode:		.byte 1		;[1] normal status data

	PktState:		.byte 1		;[1]
	PktAddress:		.byte 1		;[1] \ Packet To address
	PktCommand:		.byte 1		;[1]  | command (8 Len), (8 Cmd)
	PktData:		.byte 18	;[15] | Packet Data * n + checksum

	UartTxBuffer:		.byte 32	;[32] tx buffer
	UartHeadPtr:		.byte 1		;[1] offset to head position. updated by main code
	UartTailPtr:		.byte 1		;[1] offset to tail position. updated by INT

.CSEG

;----------------------------------------------------------------------
SerPktInit:
;purpose
;	to set up system at startup
;in
;out
;uses
;	al (register r16-31)
;Note: Ram is cleared by main

	ldi	al,0xFF
	sts	PktState,al
	sts	SerGroup,al
	ldi	al,0
	sts	SerAddress,al

	ldi	al,63				;baud = 19200 
	rcall	SerPktSetBaud			;al = baud

	ldi	al,(1<<RXCIE) | (1<<RXEN) ;
	out 	UCR,al				;UART Control Register

	ret;

;----------------------------------------------------------------------
SerPktMain:
;purpose
;	called from the main loop to process packet at 1.9K
;in
;out
;uses
;	ax,bx,cx

	MacrojmpSelectDisabled SerPktNext	;check select pin state


	lds	al,PktState	;Packet state
	cpi	al,20 
	brsh	SerPktDone	;no packet
	
	cpi	al,3 
	brlo	SerPktDone	;packet is not large enough
	
	lds	ah,PktCommand	;Command Code
	swap	ah		;swap nibbles
	andi	ah,0xF		;len only
	subi	ah,-3		;reqired bytes
	cp	al,ah		;
	brlo	SerPktDone	;not enough bytes for this packet lenght

	;we have a packet

	lds	cl,PktAddress	;this packet's address

	lds	bl,SerAddress	;my address
	cp	cl,bl
	breq	SerPktForMe	;

	lds	bh,SerGroup	;my group code
	ori	bh,0x80		;group codes must be > 127
	cp	cl,bh
	breq	SerPktForMe

	cpi	cl,0xFF		;255 is alwayse acepted
	breq	SerPktForMe	;

	rjmp	SerPktNext
SerPktForMe:			;this ones for me

	lds	ah,PktCommand	;Command Code
	andi	ah,0xF		;command only

	;16

	cpi	ah,0x0F
	breq	PktCmd0F	;hard reset does not care about select pins

	cpi	ah,0x0A
	breq	PktCmd0A	; Set Baud rate does not care about select pins


	rjmp PktDecodeThisPacket ;ah=command

SerPktNext:
	ldi	al,0xFF		;clear packet
	sts	PktState,al	;Packet state
SerPktDone:	

	rcall SerSendBytes
	ret;
;----------------------------------------------------------
PktCmd0F:				;--> hard reset
	rjmp	reset

;	rcall	SerPktInit
;	MacroSetSelectDisabled	
;	rjmp 	SerPktNext;

PktCmd0A:				;-->Set Baud rate [Baud byte] Group command only. No reply
	lds	al,PktData
	rcall	SerPktSetBaud		;set the baudrate. al=code
	rjmp	SerPktNext


PktDecodeThisPacket:
		
	cpi	ah,0x01
	brlo	PktCmd00	; Reset position
	breq	PktCmd01	; Set Address
	cpi	ah,0x03
	brlo	PktCmd02	; Set Status Type
	breq	PktCmd03	; Get custom Status
	cpi	ah,0x05
	brlo	PktCmd04	; Load trajectory
	breq	PktCmd05	; Start Motion
	cpi	ah,0x07
	brlo	PktCmd06	; Set Gain
	breq	PktCmd07	; Stop motor
	cpi	ah,0x09
	brlo	PktCmd08	; I/O control [1] (not used)
	breq	PktCmd09	; Set Home mode [1] (not used)
	cpi	ah,0x0B
	brlo	PktCmd0A	; Set Baud rate
	breq	PktCmd0B	; Clear Status Sticky bits
	cpi	ah,0x0D
	brlo	PktCmd0C	; Save current position as home
	breq	PktCmd0D	; Debug memory read/ write

;	cpi	ah,0x0E
;	breq	PktCmd0E	; status
	rjmp SerPktSendStatus

;---
PktCmd00: 				;-->Reset position
	rcall	QuadClearCounter
	rjmp	SerPktSendStatus
;---
PktCmd01: 				;-->Set Address [Address][group]	
	MacroSetSelectActive		;hardware.h
	lds	al,PktData
	sts	SerAddress,al
	lds	al,PktData+1
	sts	SerGroup,al
	rjmp	SerPktSendStatus
;---
PktCmd02: 				;-->Set Status Type code [x]
	lds	ah,PktData
	sts	SerStatusCode,ah
	rjmp	SerPktSendStatus
;---
PktCmd03:				;-->Get Custom Status [x] 
	lds	ah,PktData
	rjmp	SerPktSendCustomStatus 	;ah = status type code
;--
PktCmd04:				;-->Load trajectory [Control][0 to 13]

	rjmp	SerPktSendStatus
;---
PktCmd05:				;-->Start Motion

	rjmp	SerPktSendStatus
;---
PktCmd06:				;-->Set Gain [14]

	rjmp	SerPktSendStatus
;---
PktCmd07:				;-->Stop motor [control] [0 or 4]

	rjmp	SerPktSendStatus
;---
PktCmd08:				;-->I/O control [1] (not used)
	rjmp	SerPktSendStatus
;---
PktCmd09:				;-->Set Home mode [1] (not used)
	rjmp	SerPktSendStatus
;---

;---
PktCmd0B:				;-->Clear Status Sticky bits 
	;andi status
	rjmp	SerPktSendStatus
;---
PktCmd0C:				;-->Save current position as home
	rjmp	SerPktSendStatus
;---
PktCmd0D:				;-->Debug memory read/ write
	;	
	rjmp	SerPktNext
;---

SerPktSendStatus:
	lds	ah,SerStatusCode
SerPktSendCustomStatus:		;ah = status code

	;was this a broadcast?
	;if so am i a group leader?
	lds	cl,PktAddress	;this packet's address
	cpi	cl,128
	brsh	PktStatNoStat	;no status on broadcast packets

	clr	bl			;
	lds	al,Ram_Status		;Status byte

	rcall	SerSendByte		;al=byte, bl+=al

	sbrs	ah,0
	rjmp	PktStatSendPos
	lds	al,Ram_StatusPos+3	;sent low byte first
	rcall	SerSendByte		;al=byte, bl+=al
	lds	al,Ram_StatusPos+2	;
	rcall	SerSendByte		;al=byte, bl+=al
	lds	al,Ram_StatusPos+1	;
	rcall	SerSendByte		;al=byte, bl+=al
	lds	al,Ram_StatusPos+0	;H
	rcall	SerSendByte		;al=byte, bl+=al
PktStatSendPos:

	sbrs	ah,1
	rjmp	PktStatSendAD
	lds	al,Ram_StatusAD
	rcall	SerSendByte		;al=byte, bl+=al
PktStatSendAD:

	sbrs	ah,2
	rjmp	PktStatSendVel
	lds	al,Ram_StatusVel+1	;L
	rcall	SerSendByte		;al=byte, bl+=al
	lds	al,Ram_StatusVel	;H
	rcall	SerSendByte		;al=byte, bl+=al
PktStatSendVel:

	sbrs	ah,3
	rjmp	PktStatSendAux
	lds	al,Ram_StatusAux
	rcall	SerSendByte		;al=byte, bl+=al
PktStatSendAux:

	sbrs	ah,4
	rjmp	PktStatSendHome
	lds	al,Ram_StatusHome+3	;sent low byte first
	rcall	SerSendByte		;al=byte, bl+=al
	lds	al,Ram_StatusHome+2	;
	rcall	SerSendByte		;al=byte, bl+=al
	lds	al,Ram_StatusHome+1	;
	rcall	SerSendByte		;al=byte, bl+=al
	lds	al,Ram_StatusHome+0	;H
	rcall	SerSendByte		;al=byte, bl+=al
PktStatSendHome:

	sbrs	ah,5
	rjmp	PktStatSendDevType
	lds	al,Ram_StatusDevType+1	;L
	rcall	SerSendByte		;al=byte, bl+=al
	lds	al,Ram_StatusDevType	;H
	rcall	SerSendByte		;al=byte, bl+=al
PktStatSendDevType:

	mov	al,bl		;checksum
	rcall	SerSendByte		;al=byte, bl+=al

PktStatNoStat:
	rjmp SerPktNext


;----------------------------------------------------------------
SerPktSetBaud:
;Purpose: To set the baud rate based on devisor
;in: al = 
;		 al		9.216Mhz
;	9600   = 129 120  	59
;	19200  = 63 60		29 	default
;	38400  = 30		14
;	57600  = 20		9
;	115200 = 10		4
;from
; Baud = 20Mhz / (16 * (x+1)) = 9615
;To
; Baud = Clk / (16 * (n+1))
;
	cpi	al,10
	breq	SerSetBaud115200
	ldi	al,29 			;9.2 mhz 19200
	rjmp	SerSetBaud
SerSetBaud115200:
	ldi	al,4 			;9.2 mhz 115200

SerSetBaud:
	out	UBRR,al

	ret;
;--------------------------------------------------------------------------------------
;SerSendString:
;purpose
;	to send a string
;in
;	Z -> string
;uses
;	Y,Z

SerSendByte:
;purpose
;	to add a byte to the transit buffer
;in
;	al = byte
;	bl
;out
;	bl = bl + al for checksum	
;uses
;	Y
	add	bl,al			;chk

	clr	yh				;1
	lds	yl,UartHeadPtr			;1,2 get head pointer
	subi	yl,low((-UartTxBuffer))		;1   offset
	sbci 	yh,high((-UartTxBuffer))	;1 
	st	y,al				;1   store

	lds	yl,UartHeadPtr			;1,2 get head pointer
	inc	yl				;1
	cpi	yl,32				;1
	brlo	SerSendByteHeadRoll		;1
	clr	yl				;1   roll head pointer
SerSendByteHeadRoll:
	sts	UartHeadPtr,YL			;1,2 save head pointer

	;enable tx

	ret;
;--------------
SerSendBytes: ; test !! This will eventually be done on an interrupt

	
	lds	al,UartHeadPtr
	clr	yh				;1
	lds	yl,UartTailPtr			;1,2 get head pointer
	cp	al,yl
	breq	SerSendByteTailEmpty		;empty

	in	ah,USR
	sbrs	ah,UDRE
	ret					;tx busy

	in	al,UCR
	ori	al,(1<<TXEN)			;enable transmiter
	out 	UCR,al				;UART Control Register

	subi	yl,low(-(UartTxBuffer))		;1   offset
	sbci 	yh,high(-(UartTxBuffer))	;1 
	ld	al,y				;1   
	out	UDR,al

	sbi	USR,TXC				;clear complete flag

	lds	yl,UartTailPtr			;1,2 get head pointer
	inc	yl				;1
	cpi	yl,32				;1
	brlo	SerSendByteTailRoll		;1
	clr	yl				;1   roll head pointer
SerSendByteTailRoll:
	sts	UartTailPtr,YL			;1,2 get head pointer
SerSendByteTailDone:
	ret

SerSendByteTailEmpty:

	in	al,USR
	sbrs	al,TXC
	ret

	in	al,UCR
	andi	al,~(1<<TXEN)			;dis enable transmiter
	out 	UCR,al				;UART Control Register

	ret;					;tx busy


;**************************************************************************
UART_RXC:	;Uart Rx Complete received byte

;purpose
;	To decode incoming data into packets
;in
;out
;uses
;	ax
;	Z	
;Cycles
	in	REG_SerRXCsreg, sreg
	push 	al	
	in	al,UDR			; Get data and clear interrupt flag
	sei				; Enable interrupts!
	
	push	ah	
	pushZ
;---
	lds	ah,PktState		;ff=wait for start, else size of packet
	cpi	ah,18			;limit received packet size
	brlo	URxcPktStateN
	cpi	al,0xAA			;is it the start byte?
	brne	URxcPktStateDone	
	ldi	ah,0			;start packet
	rjmp	URxcPktStateDone
URxcPktStateN:
	clr	zh
	mov	zl,ah
	subi	zl,low((-PktAddress))	;offset
	sbci 	Zh,high((-PktAddress))
	st	Z,al			;store

	inc	ah 			;next
URxcPktStateDone:
	sts	PktState,ah		;save state
;---
	PopZ
	pop	ah
	pop	al
	out	sreg, REG_SerRXCsreg

	reti		;
;------------------------------------------------------------------------
UART_DRE: 	;Uart Tx Data register empty
;purpose
;	To handle Tx Buffer
;in
;out
;uses
;
;Cycles

	in	REG_SerDREsreg,sreg

	push 	al
	push	ah

	;out	UDR,al			;send data and clear interupt flag
	sei				; Enable interrupts! oops? reinterupt?
 
	;in	al,UCR
	
	pop	ah
	pop	al
	out	sreg, REG_SerRXCsreg

	reti;
;------------------------------------------------------------------------

UART_TXC: 	;Uart Tx Complete. 

	in	REG_SerTXCsreg,sreg
	sei				; Enable interrupts!
	push 	al
	push	ah


	;tristate port

	pop	ah
	pop	al
	out	sreg, REG_SerRXCsreg


	RETI;				;1,4

